import yaml from "yaml";
import path from "path";
import fs from 'fs';

import viewStore from "./view_store.js";
import commandGateway from "./command_gateway.js";
import viewGateway from "./view_gateway.js";
import factGateway from "./fact_gateway.js";
import command from "./command.js";
import projection from "./projection.js";
import eventStore from "./event_store.js";
import job from "./job.js";
import fact from "./fact.js";
import viewComposite from "./view_composite.js";

const checkInPath = "_check-in";
const checkInSchedule = "0 * * * *";

const steps = ({
  region,
  domain,
  name,
  actions,
  events,
  publicKeyUrl,
  project,
  network,
  timeout,
  mongodbUser,
  mongodbHost,
  mongodbProtocol,
  memory,
  envUriSpecifier,
  containerRegistery,
  mainContainerName,
  baseNetwork,
  dnsZone,
  service,
  procedure,
  host,
  env,
  context,
  redisIp,
  localBaseNetwork,
  localNetwork,
  bootstrapContext,
  operationHash,
  operationName,
  envVars,
  devEnvVars,
  rolesBucket,
  secretBucket,
  secretBucketKeyLocation,
  secretBucketKeyRing,
  blockSchedule,
  imageExtension,
  runUnitTests,
  runBaseUnitTests,
  runIntegrationTests,
  runBaseIntegrationTests,
  computeUrlId,
  strict,
  dependencyKeyEnvironmentVariables,
}) => {
  const serviceName = `${region}-${operationName}-${operationHash}`;
  const uri = `${operationHash}.${region}.${envUriSpecifier}${network}`;
  switch (procedure) {
    case "view-store":
      return viewStore({
        imageExtension,
        region,
        name,
        project,
        network,
        context,
        bootstrapContext,
        localBaseNetwork,
        localNetwork,
        memory,
        host,
        computeUrlId,
        envUriSpecifier,
        dependencyKeyEnvironmentVariables,
        containerRegistery,
        mainContainerName,
        dnsZone,
        timeout,
        procedure,
        operationHash,
        operationName,
        serviceName,
        baseNetwork,
        redisIp,
        env,
        uri,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        mongodbUser,
        mongodbHost,
        mongodbProtocol,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "view-composite":
      return viewComposite({
        imageExtension,
        region,
        domain,
        service,
        name,
        project,
        network,
        context,
        redisIp,
        localBaseNetwork,
        localNetwork,
        host,
        memory,
        computeUrlId,
        envUriSpecifier,
        dependencyKeyEnvironmentVariables,
        containerRegistery,
        mainContainerName,
        dnsZone,
        procedure,
        operationHash,
        operationName,
        serviceName,
        baseNetwork,
        timeout,
        env,
        uri,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "event-store":
      return eventStore({
        imageExtension,
        domain,
        actions,
        region,
        project,
        dnsZone,
        service,
        procedure,
        network,
        host,
        redisIp,
        computeUrlId,
        localBaseNetwork,
        localNetwork,
        envUriSpecifier,
        baseNetwork,
        memory,
        env,
        serviceName,
        dependencyKeyEnvironmentVariables,
        containerRegistery,
        mainContainerName,
        operationHash,
        timeout,
        blockSchedule,
        uri,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        mongodbUser,
        mongodbHost,
        mongodbProtocol,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "projection":
      return projection({
        imageExtension,
        region,
        name,
        events,
        project,
        network,
        host,
        dependencyKeyEnvironmentVariables,
        envUriSpecifier,
        localBaseNetwork,
        localNetwork,
        computeUrlId,
        redisIp,
        containerRegistery,
        mainContainerName,
        baseNetwork,
        timeout,
        dnsZone,
        context,
        memory,
        procedure,
        operationName,
        operationHash,
        env,
        serviceName,
        envVars,
        devEnvVars,
        uri,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "command":
      return command({
        imageExtension,
        name,
        region,
        domain,
        project,
        host,
        network,
        memory,
        computeUrlId,
        localBaseNetwork,
        localNetwork,
        baseNetwork,
        dependencyKeyEnvironmentVariables,
        mainContainerName,
        redisIp,
        envUriSpecifier,
        containerRegistery,
        operationHash,
        dnsZone,
        service,
        procedure,
        timeout,
        env,
        serviceName,
        uri,
        envVars,
        devEnvVars,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "job":
      return job({
        imageExtension,
        name,
        domain,
        region,
        project,
        network,
        localBaseNetwork,
        localNetwork,
        host,
        envUriSpecifier,
        containerRegistery,
        mainContainerName,
        dependencyKeyEnvironmentVariables,
        baseNetwork,
        dnsZone,
        service,
        procedure,
        timeout,
        memory,
        computeUrlId,
        operationHash,
        env,
        serviceName,
        uri,
        redisIp,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        envVars,
        devEnvVars,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "fact":
      return fact({
        imageExtension,
        name,
        domain,
        region,
        project,
        network,
        localBaseNetwork,
        localNetwork,
        host,
        envUriSpecifier,
        containerRegistery,
        mainContainerName,
        dependencyKeyEnvironmentVariables,
        baseNetwork,
        dnsZone,
        service,
        procedure,
        timeout,
        memory,
        computeUrlId,
        operationHash,
        env,
        serviceName,
        redisIp,
        uri,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        envVars,
        devEnvVars,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "command-gateway":
      return commandGateway({
        imageExtension,
        publicKeyUrl,
        region,
        project,
        envUriSpecifier,
        containerRegistery,
        mainContainerName,
        dependencyKeyEnvironmentVariables,
        dnsZone,
        localBaseNetwork,
        localNetwork,
        memory,
        baseNetwork,
        env,
        host,
        envVars,
        redisIp,
        timeout,
        devEnvVars,
        computeUrlId,
        service,
        domain,
        procedure,
        network,
        operationHash,
        operationName,
        serviceName,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "view-gateway":
      return viewGateway({
        imageExtension,
        publicKeyUrl,
        region,
        project,
        context,
        host,
        envUriSpecifier,
        containerRegistery,
        localBaseNetwork,
        localNetwork,
        mainContainerName,
        dependencyKeyEnvironmentVariables,
        computeUrlId,
        dnsZone,
        baseNetwork,
        redisIp,
        memory,
        timeout,
        env,
        procedure,
        network,
        operationHash,
        operationName,
        serviceName,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
    case "fact-gateway":
      return factGateway({
        imageExtension,
        publicKeyUrl,
        region,
        project,
        envUriSpecifier,
        containerRegistery,
        mainContainerName,
        baseNetwork,
        computeUrlId,
        localBaseNetwork,
        localNetwork,
        dependencyKeyEnvironmentVariables,
        dnsZone,
        timeout,
        host,
        redisIp,
        memory,
        env,
        service,
        domain,
        procedure,
        network,
        operationHash,
        operationName,
        serviceName,
        rolesBucket,
        secretBucket,
        secretBucketKeyLocation,
        secretBucketKeyRing,
        runUnitTests,
        runBaseUnitTests,
        runIntegrationTests,
        runBaseIntegrationTests,
        strict,
        checkInPath,
        checkInSchedule,
      });
  }
};

const imageExtension = ({ service, context, domain, name, procedure }) => {
  switch (procedure) {
    case "view-store":
    case "projection":
    case "view-composite":
      return `${context ? `${context}.` : ""}${name}`;
    case "event-store":
    case "command-gateway":
      return `${service}.${domain}`;
    case "view-gateway":
      return context || "";
    case "fact-gateway":
      if (service) {
        if (domain) return `${service}.${domain}`;
        return service;
      }
      return "";
    case "command":
      return `${service}.${domain}.${name}`;
    case "job":
    case "fact":
      return `${service ? `${service}.` : ""}${
        domain ? `${domain}.` : ""
      }${name}`;
    default:
      return "";
  }
};

export default ({
  workingDir,
  region,
  domain,
  actions,
  events,
  publicKeyUrl,
  name,
  localBaseNetwork,
  localNetwork,
  project,
  network,
  procedure,
  memory,
  host,
  redisIp,
  context,
  bootstrapContext,
  envUriSpecifier,
  baseNetwork,
  containerRegistery,
  mainContainerName,
  dependencyKeyEnvironmentVariables,
  blockSchedule,
  mongodbUser,
  mongodbHost,
  mongodbProtocol,
  dnsZone,
  service,
  operationHash,
  operationName,
  envVars,
  devEnvVars,
  timeout,
  env,
  rolesBucket,
  secretBucket,
  secretBucketKeyLocation,
  secretBucketKeyRing,
  computeUrlId,
  strict,
}) => {
  const buildPath = path.resolve(workingDir, "build.yaml");

  const i = imageExtension({
    procedure,
    name,
    domain,
    service,
    context,
    bootstrapContext,
  });

  const runUnitTests = fs.existsSync(path.resolve(workingDir, "test/unit"));
  const runBaseUnitTests = fs.existsSync(
    path.resolve(workingDir, "base_test/unit")
  );
  const runIntegrationTests = fs.existsSync(
    path.resolve(workingDir, "test/integration")
  );
  const runBaseIntegrationTests = fs.existsSync(
    path.resolve(workingDir, "base_test/integration")
  );

  const build = {
    steps: steps({
      imageExtension: i,
      region,
      domain,
      name,
      localBaseNetwork,
      localNetwork,
      actions,
      events,
      publicKeyUrl,
      project,
      network,
      context,
      bootstrapContext,
      memory,
      procedure,
      envVars,
      devEnvVars,
      envUriSpecifier,
      dependencyKeyEnvironmentVariables,
      containerRegistery,
      mainContainerName,
      dnsZone,
      service,
      timeout,
      env,
      host,
      redisIp,
      blockSchedule,
      baseNetwork,
      computeUrlId,
      operationHash,
      operationName,
      mongodbHost,
      mongodbUser,
      mongodbProtocol,
      rolesBucket,
      secretBucket,
      secretBucketKeyLocation,
      secretBucketKeyRing,
      runUnitTests,
      runBaseUnitTests,
      runIntegrationTests,
      runBaseIntegrationTests,
      strict,
    }),
  };

  fs.writeFileSync(buildPath, yaml.stringify(build));
};
